@@ -2,100 +2,119 @@
 //! code that generalizes over a pair of fields.
 
 use core::mem::size_of;
+
 use static_assertions::const_assert;
-use std::assert;
-use std::convert::TryInto;
-use std::marker::PhantomData;
-use subtle::{Choice, CtOption};
+use subtle::{Choice, ConditionallySelectable, CtOption};
 
 use super::Group;
 
-use std::io::{self, Read, Write};
+use core::assert;
+
+#[cfg(feature = "sqrt-table")]
+use alloc::{boxed::Box, vec::Vec};
+#[cfg(feature = "sqrt-table")]
+use core::{convert::TryInto, marker::PhantomData};
 
 const_assert!(size_of::<usize>() >= 4);
 
+/// A trait that exposes additional operations related to calculating square roots of
+/// prime-order finite fields.
+pub trait SqrtRatio: ff::PrimeField {
+    /// The value $(T-1)/2$ such that $2^S \cdot T = p - 1$ with $T$ odd.
+    const T_MINUS1_OVER2: [u64; 4];
+
+    /// Raise this field element to the power [`Self::T_MINUS1_OVER2`].
+    ///
+    /// Field implementations may override this to use an efficient addition chain.
+    fn pow_by_t_minus1_over2(&self) -> Self {
+        ff::Field::pow_vartime(&self, &Self::T_MINUS1_OVER2)
+    }
+
+    /// Gets the lower 32 bits of this field element when expressed
+    /// canonically.
+    fn get_lower_32(&self) -> u32;
+
+    /// Computes:
+    ///
+    /// - $(\textsf{true}, \sqrt{\textsf{num}/\textsf{div}})$, if $\textsf{num}$ and
+    ///   $\textsf{div}$ are nonzero and $\textsf{num}/\textsf{div}$ is a square in the
+    ///   field;
+    /// - $(\textsf{true}, 0)$, if $\textsf{num}$ is zero;
+    /// - $(\textsf{false}, 0)$, if $\textsf{num}$ is nonzero and $\textsf{div}$ is zero;
+    /// - $(\textsf{false}, \sqrt{G_S \cdot \textsf{num}/\textsf{div}})$, if
+    ///   $\textsf{num}$ and $\textsf{div}$ are nonzero and $\textsf{num}/\textsf{div}$ is
+    ///   a nonsquare in the field;
+    ///
+    /// where $G_S$ is a non-square.
+    ///
+    /// For `pasta_curves`, $G_S$ is currently [`ff::PrimeField::root_of_unity`], a
+    /// generator of the order $2^S$ subgroup. Users of this crate should not rely on this
+    /// generator being fixed; it may be changed in future crate versions to simplify the
+    /// implementation of the SSWU hash-to-curve algorithm.
+    ///
+    /// The choice of root from sqrt is unspecified.
+    fn sqrt_ratio(num: &Self, div: &Self) -> (Choice, Self) {
+        // General implementation:
+        //
+        // a = num * inv0(div)
+        //   = {    0    if div is zero
+        //     { num/div otherwise
+        //
+        // b = G_S * a
+        //   = {      0      if div is zero
+        //     { G_S*num/div otherwise
+        //
+        // Since G_S is non-square, a and b are either both zero (and both square), or
+        // only one of them is square. We can therefore choose the square root to return
+        // based on whether a is square, but for the boolean output we need to handle the
+        // num != 0 && div == 0 case specifically.
+
+        let a = div.invert().unwrap_or_else(Self::zero) * num;
+        let b = a * Self::root_of_unity();
+        let sqrt_a = a.sqrt();
+        let sqrt_b = b.sqrt();
+
+        let num_is_zero = num.is_zero();
+        let div_is_zero = div.is_zero();
+        let is_square = sqrt_a.is_some();
+        let is_nonsquare = sqrt_b.is_some();
+        assert!(bool::from(
+            num_is_zero | div_is_zero | (is_square ^ is_nonsquare)
+        ));
+
+        (
+            is_square & !(!num_is_zero & div_is_zero),
+            CtOption::conditional_select(&sqrt_b, &sqrt_a, is_square).unwrap(),
+        )
+    }
+
+    /// Equivalent to `Self::sqrt_ratio(self, one())`.
+    fn sqrt_alt(&self) -> (Choice, Self) {
+        Self::sqrt_ratio(self, &Self::one())
+    }
+}
+
 /// This trait is a common interface for dealing with elements of a finite
 /// field.
-pub trait FieldExt: ff::PrimeField + From<bool> + Ord + Group<Scalar = Self> {
+pub trait FieldExt: SqrtRatio + From<bool> + Ord + Group<Scalar = Self> {
     /// Modulus of the field written as a string for display purposes
     const MODULUS: &'static str;
 
-    /// Generator of the $2^S$ multiplicative subgroup
-    const ROOT_OF_UNITY: Self;
-
-    /// Inverse of `ROOT_OF_UNITY`
+    /// Inverse of `PrimeField::root_of_unity()`
     const ROOT_OF_UNITY_INV: Self;
 
-    /// The value $(T-1)/2$ such that $2^S \cdot T = p - 1$ with $T$ odd.
-    const T_MINUS1_OVER2: [u64; 4];
-
     /// Generator of the $t-order$ multiplicative subgroup
     const DELTA: Self;
 
     /// Inverse of $2$ in the field.
     const TWO_INV: Self;
 
-    /// Ideally the smallest prime $\alpha$ such that gcd($p - 1$, $\alpha$) = $1$
-    const RESCUE_ALPHA: u64;
-
-    /// $RESCUE_INVALPHA \cdot RESCUE_ALPHA = 1 \mod p - 1$ such that
-    /// `(a^RESCUE_ALPHA)^RESCUE_INVALPHA = a`.
-    const RESCUE_INVALPHA: [u64; 4];
-
     /// Element of multiplicative order $3$.
     const ZETA: Self;
 
-    /// Computes:
-    ///
-    /// * (true,  sqrt(num/div)),                 if num and div are nonzero and num/div is a square in the field;
-    /// * (true,  0),                             if num is zero;
-    /// * (false, 0),                             if num is nonzero and div is zero;
-    /// * (false, sqrt(ROOT_OF_UNITY * num/div)), if num and div are nonzero and num/div is a nonsquare in the field;
-    ///
-    /// where ROOT_OF_UNITY is a generator of the order 2^n subgroup (and therefore a nonsquare).
-    ///
-    /// The choice of root from sqrt is unspecified.
-    fn sqrt_ratio(num: &Self, div: &Self) -> (Choice, Self);
-
-    /// Equivalent to sqrt_ratio(self, one()).
-    fn sqrt_alt(&self) -> (Choice, Self) {
-        Self::sqrt_ratio(self, &Self::one())
-    }
-
-    /// This computes a random element of the field using system randomness.
-    fn rand() -> Self {
-        Self::random(rand::rngs::OsRng)
-    }
-
-    /// Obtains a field element congruent to the integer `v`.
-    fn from_u64(v: u64) -> Self;
-
     /// Obtains a field element congruent to the integer `v`.
     fn from_u128(v: u128) -> Self;
 
-    /// Converts this field element to its normalized, little endian byte
-    /// representation.
-    fn to_bytes(&self) -> [u8; 32];
-
-    /// Writes this element in its normalized, little endian form into a buffer.
-    fn write<W: Write>(&self, writer: &mut W) -> io::Result<()> {
-        let compressed = self.to_bytes();
-        writer.write_all(&compressed[..])
-    }
-
-    /// Attempts to obtain a field element from its normalized, little endian
-    /// byte representation.
-    fn from_bytes(bytes: &[u8; 32]) -> CtOption<Self>;
-
-    /// Reads a normalized, little endian represented field element from a
-    /// buffer.
-    fn read<R: Read>(reader: &mut R) -> io::Result<Self> {
-        let mut compressed = [0u8; 32];
-        reader.read_exact(&mut compressed[..])?;
-        Option::from(Self::from_bytes(&compressed))
-            .ok_or_else(|| io::Error::new(io::ErrorKind::Other, "invalid point encoding in proof"))
-    }
-
     /// Obtains a field element that is congruent to the provided little endian
     /// byte representation of an integer.
     fn from_bytes_wide(bytes: &[u8; 64]) -> Self;
@@ -118,19 +137,62 @@ pub trait FieldExt: ff::PrimeField + From<bool> + Ord + Group<Scalar = Self> {
     /// Gets the lower 128 bits of this field element when expressed
     /// canonically.
     fn get_lower_128(&self) -> u128;
+}
 
-    /// Gets the lower 32 bits of this field element when expressed
-    /// canonically.
-    fn get_lower_32(&self) -> u32;
+/// Tonelli–Shanks' square-root algorithm for `p mod 16 = 1`.
+///
+/// https://eprint.iacr.org/2012/685.pdf (page 12, algorithm 5)
+///
+/// `tm1d2` should be set to `(t - 1) // 2`, where `t = (modulus - 1) >> F::S`.
+#[cfg(not(feature = "sqrt-table"))]
+#[cfg_attr(docsrs, doc(cfg(not(feature = "sqrt-table"))))]
+pub(crate) fn sqrt_tonelli_shanks<F: ff::PrimeField, S: AsRef<[u64]>>(
+    f: &F,
+    tm1d2: S,
+) -> CtOption<F> {
+    use subtle::ConstantTimeEq;
 
-    /// Raise this field element to the power T_MINUS1_OVER2.
-    /// Field implementations may override this to use an efficient addition chain.
-    fn pow_by_t_minus1_over2(&self) -> Self {
-        ff::Field::pow_vartime(&self, &Self::T_MINUS1_OVER2)
+    // w = self^((t - 1) // 2)
+    let w = f.pow_vartime(tm1d2);
+
+    let mut v = F::S;
+    let mut x = w * f;
+    let mut b = x * w;
+
+    // Initialize z as the 2^S root of unity.
+    let mut z = F::root_of_unity();
+
+    for max_v in (1..=F::S).rev() {
+        let mut k = 1;
+        let mut tmp = b.square();
+        let mut j_less_than_v: Choice = 1.into();
+
+        for j in 2..max_v {
+            let tmp_is_one = tmp.ct_eq(&F::one());
+            let squared = F::conditional_select(&tmp, &z, tmp_is_one).square();
+            tmp = F::conditional_select(&squared, &tmp, tmp_is_one);
+            let new_z = F::conditional_select(&z, &squared, tmp_is_one);
+            j_less_than_v &= !j.ct_eq(&v);
+            k = u32::conditional_select(&j, &k, tmp_is_one);
+            z = F::conditional_select(&z, &new_z, j_less_than_v);
+        }
+
+        let result = x * z;
+        x = F::conditional_select(&result, &x, b.ct_eq(&F::one()));
+        z = z.square();
+        b *= z;
+        v = k;
     }
+
+    CtOption::new(
+        x,
+        (x * x).ct_eq(f), // Only return Some if it's the square root.
+    )
 }
 
 /// Parameters for a perfect hash function used in square root computation.
+#[cfg(feature = "sqrt-table")]
+#[cfg_attr(docsrs, doc(cfg(feature = "sqrt-table")))]
 #[derive(Debug)]
 struct SqrtHasher<F: FieldExt> {
     hash_xor: u32,
@@ -138,6 +200,7 @@ struct SqrtHasher<F: FieldExt> {
     marker: PhantomData<F>,
 }
 
+#[cfg(feature = "sqrt-table")]
 impl<F: FieldExt> SqrtHasher<F> {
     /// Returns a perfect hash of x for use with SqrtTables::inv.
     fn hash(&self, x: &F) -> usize {
@@ -150,6 +213,8 @@ impl<F: FieldExt> SqrtHasher<F> {
 }
 
 /// Tables used for square root computation.
+#[cfg(feature = "sqrt-table")]
+#[cfg_attr(docsrs, doc(cfg(feature = "sqrt-table")))]
 #[derive(Debug)]
 pub struct SqrtTables<F: FieldExt> {
     hasher: SqrtHasher<F>,
@@ -160,16 +225,19 @@ pub struct SqrtTables<F: FieldExt> {
     g3: Box<[F; 129]>,
 }
 
+#[cfg(feature = "sqrt-table")]
 impl<F: FieldExt> SqrtTables<F> {
     /// Build tables given parameters for the perfect hash.
     pub fn new(hash_xor: u32, hash_mod: usize) -> Self {
+        use alloc::vec;
+
         let hasher = SqrtHasher {
             hash_xor,
             hash_mod,
             marker: PhantomData,
         };
 
-        let mut gtab = (0..4).scan(F::ROOT_OF_UNITY, |gi, _| {
+        let mut gtab = (0..4).scan(F::root_of_unity(), |gi, _| {
             // gi == ROOT_OF_UNITY^(256^i)
             let gtab_i: Vec<F> = (0..256)
                 .scan(F::one(), |acc, _| {
@@ -267,7 +335,7 @@ impl<F: FieldExt> SqrtTables<F> {
 
         let sqdiv = res.square() * div;
         let is_square = (sqdiv - num).is_zero();
-        let is_nonsquare = (sqdiv - F::ROOT_OF_UNITY * num).is_zero();
+        let is_nonsquare = (sqdiv - F::root_of_unity() * num).is_zero();
         assert!(bool::from(
             num.is_zero() | div.is_zero() | (is_square ^ is_nonsquare)
         ));
@@ -284,7 +352,7 @@ impl<F: FieldExt> SqrtTables<F> {
 
         let sq = res.square();
         let is_square = (sq - u).is_zero();
-        let is_nonsquare = (sq - F::ROOT_OF_UNITY * u).is_zero();
+        let is_nonsquare = (sq - F::root_of_unity() * u).is_zero();
         assert!(bool::from(u.is_zero() | (is_square ^ is_nonsquare)));
 
         (is_square, res)
