@@ -2,10 +2,14 @@
 //! groups.
 
 use core::cmp;
-use core::fmt::Debug;
+use core::fmt;
 use core::iter::Sum;
 use core::ops::{Add, Mul, Neg, Sub};
-use ff::Field;
+
+#[cfg(feature = "alloc")]
+use alloc::boxed::Box;
+
+use ff::{Field, PrimeField};
 use group::{
     cofactor::{CofactorCurve, CofactorGroup},
     prime::{PrimeCurve, PrimeCurveAffine, PrimeGroup},
@@ -15,7 +19,10 @@ use rand::RngCore;
 use subtle::{Choice, ConditionallySelectable, ConstantTimeEq, CtOption};
 
 use super::{Fp, Fq};
-use crate::arithmetic::{Coordinates, CurveAffine, CurveExt, FieldExt, Group};
+use crate::arithmetic::Group;
+
+#[cfg(feature = "alloc")]
+use crate::arithmetic::{Coordinates, CurveAffine, CurveExt, FieldExt};
 
 macro_rules! new_curve_impl {
     (($($privacy:tt)*), $name:ident, $name_affine:ident, $iso:ident, $base:ident, $scalar:ident,
@@ -47,8 +54,8 @@ macro_rules! new_curve_impl {
             infinity: Choice,
         }
 
-        impl std::fmt::Debug for $name_affine {
-            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
+        impl fmt::Debug for $name_affine {
+            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> Result<(), fmt::Error> {
                 if self.infinity.into() {
                     write!(f, "Infinity")
                 } else {
@@ -68,7 +75,7 @@ macro_rules! new_curve_impl {
                     let x3 = x.square() * x;
                     let y = (x3 + $name::curve_constant_b()).sqrt();
                     if let Some(y) = Option::<$base>::from(y) {
-                        let sign = y.to_bytes()[0] & 1;
+                        let sign = y.is_odd().unwrap_u8();
                         let y = if ysign ^ sign == 0 { y } else { -y };
 
                         let p = $name_affine {
@@ -96,6 +103,8 @@ macro_rules! new_curve_impl {
             }
         }
 
+        #[cfg(feature = "alloc")]
+        #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
         impl group::WnafGroup for $name {
             fn recommended_wnaf_for_num_scalars(num_scalars: usize) -> usize {
                 // Copied from bls12_381::g1, should be updated.
@@ -115,6 +124,8 @@ macro_rules! new_curve_impl {
             }
         }
 
+        #[cfg(feature = "alloc")]
+        #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
         impl CurveExt for $name {
             type ScalarExt = $scalar;
             type Base = $base;
@@ -463,9 +474,14 @@ macro_rules! new_curve_impl {
                 // multiplication, moving from most significant to least
                 // significant bit of the scalar.
                 //
-                // NOTE: We skip the leading bit because it's always unset.
+                // We don't use `PrimeFieldBits::.to_le_bits` here, because that would
+                // force users of this crate to depend on `bitvec` where they otherwise
+                // might not need to.
+                //
+                // NOTE: We skip the leading bit because it's always unset (we are turning
+                // the 32-byte repr into 256 bits, and $scalar::NUM_BITS = 255).
                 for bit in other
-                    .to_bytes()
+                    .to_repr()
                     .iter()
                     .rev()
                     .flat_map(|byte| (0..8).rev().map(move |i| Choice::from((byte >> i) & 1u8)))
@@ -574,9 +590,14 @@ macro_rules! new_curve_impl {
                 // multiplication, moving from most significant to least
                 // significant bit of the scalar.
                 //
-                // NOTE: We skip the leading bit because it's always unset.
+                // We don't use `PrimeFieldBits::.to_le_bits` here, because that would
+                // force users of this crate to depend on `bitvec` where they otherwise
+                // might not need to.
+                //
+                // NOTE: We skip the leading bit because it's always unset (we are turning
+                // the 32-byte repr into 256 bits, and $scalar::NUM_BITS = 255).
                 for bit in other
-                    .to_bytes()
+                    .to_repr()
                     .iter()
                     .rev()
                     .flat_map(|byte| (0..8).rev().map(move |i| Choice::from((byte >> i) & 1u8)))
@@ -646,11 +667,11 @@ macro_rules! new_curve_impl {
                 let ysign = Choice::from(tmp[31] >> 7);
                 tmp[31] &= 0b0111_1111;
 
-                $base::from_bytes(&tmp).and_then(|x| {
+                $base::from_repr(tmp).and_then(|x| {
                     CtOption::new(Self::identity(), x.is_zero() & (!ysign)).or_else(|| {
                         let x3 = x.square() * x;
                         (x3 + $name::curve_constant_b()).sqrt().and_then(|y| {
-                            let sign = Choice::from(y.to_bytes()[0] & 1);
+                            let sign = y.is_odd();
 
                             let y = $base::conditional_select(&y, &-y, ysign ^ sign);
 
@@ -678,14 +699,16 @@ macro_rules! new_curve_impl {
                     [0; 32]
                 } else {
                     let (x, y) = (self.x, self.y);
-                    let sign = (y.to_bytes()[0] & 1) << 7;
-                    let mut xbytes = x.to_bytes();
+                    let sign = y.is_odd().unwrap_u8() << 7;
+                    let mut xbytes = x.to_repr();
                     xbytes[31] |= sign;
                     xbytes
                 }
             }
         }
 
+        #[cfg(feature = "alloc")]
+        #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
         impl CurveAffine for $name_affine {
             type ScalarExt = $scalar;
             type Base = $base;
@@ -869,6 +892,7 @@ macro_rules! impl_projective_curve_specific {
     };
 }
 
+#[cfg(feature = "alloc")]
 macro_rules! impl_projective_curve_ext {
     ($name:ident, $iso:ident, $base:ident, special_a0_b5) => {
         fn hash_to_curve<'a>(domain_prefix: &'a str) -> Box<dyn Fn(&[u8]) -> Self + 'a> {
@@ -1089,7 +1113,7 @@ impl Ep {
         0x4000000000000000,
     ]);
 
-    /// `(F::ROOT_OF_UNITY.invert().unwrap() * z).sqrt().unwrap()`
+    /// `(F::root_of_unity().invert().unwrap() * z).sqrt().unwrap()`
     pub const THETA: Fp = Fp::from_raw([
         0xca330bcc09ac318e,
         0x51f64fc4dc888857,
@@ -1189,7 +1213,7 @@ impl Eq {
         0x4000000000000000,
     ]);
 
-    /// `(F::ROOT_OF_UNITY.invert().unwrap() * z).sqrt().unwrap()`
+    /// `(F::root_of_unity().invert().unwrap() * z).sqrt().unwrap()`
     pub const THETA: Fq = Fq::from_raw([
         0x632cae9872df1b5d,
         0x38578ccadf03ac27,
